'''
    Tests for ngshare APIs
'''

import os
import json
import base64
import hashlib
import datetime
import tempfile
import shutil
from urllib.parse import urlencode
import pytest
from tornado.httputil import url_concat

from .ngshare import (
    MyApplication,
    MyHelpers,
    MockAuth,
    RequestHandler,
    MyHelpers,
    MyRequestHandler,
)

application, db_name, storage_name = None, None, None
user, hc, bu = None, None, None


@pytest.fixture
def app():
    'Create Tornado application for testing'
    global application, db_name, storage_name
    # Create temporary location for db and storage
    if db_name is None:
        db_name = tempfile.mktemp('.db')
        storage_name = tempfile.mktemp('-ngshare-test-dir')
    application = MyApplication(
        '/api/',
        'sqlite:///' + db_name,
        storage_name,
        admin=['root'],
        debug=True,
    )
    # Mock authentication using vngshare
    MyRequestHandler.__bases__ = (MockAuth, RequestHandler, MyHelpers)
    return application


def server_communication(url, data=None, params=None, method='GET'):
    'Request a page'
    global hc, bu, user
    if user is not None:
        if method != 'POST':
            params = params if params is not None else {}
            params['user'] = user
        else:
            data = data if data is not None else {}
            data['user'] = user
    actual_url = bu + url_concat(url, params)
    if method != 'POST':
        body = None
    else:
        body = urlencode(data)
    response = yield hc.fetch(
        actual_url, method=method, body=body, raise_error=False
    )
    return response


def assert_fail(url, data=None, params=None, method='GET', msg=None):
    'Assert requesting a page is failing (with matching message)'
    response = yield from server_communication(url, data, params, method)
    assert response.code in range(400, 500)
    resp = json.loads(response.body)
    assert resp['message'] == msg
    return resp


def assert_success(url, data=None, params=None, method='GET'):
    'Assert requesting a page is success'
    response = yield from server_communication(url, data, params, method)
    assert response.code == 200
    resp = json.loads(response.body)
    assert resp['success'] == True
    return resp


@pytest.mark.gen_test
def test_health(http_client, base_url):
    'Test /healthz endpoint'
    response = yield http_client.fetch(base_url + '/healthz')
    assert response.code == 200
    assert json.loads(response.body)['success']


@pytest.mark.gen_test
def test_api_endpoint_conflict(http_client, base_url):
    'Test throwing an error if the API prefix starts with /healthz/'
    with pytest.raises(ValueError):
        MyApplication("/healthz/", 'sqlite:////tmp/ngshare.db', '/tmp/ngshare/')


@pytest.mark.gen_test
def test_home(http_client, base_url):
    'Test homepage, favicon, etc.'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'none'
    response = yield from server_communication('/api/')
    assert response.code == 200
    assert response.body.decode().startswith('<!doctype html>')
    response = yield from server_communication('/api/favicon.ico')
    assert response.code == 200
    assert response.body.startswith(b'\x89PNG')
    response = yield from server_communication('/api/masonry.min.js')
    assert response.code == 200
    assert 'Masonry' in response.body.decode()
    response = yield from server_communication('/api/random-page')
    assert response.code == 404
    assert '<h1>404 Not Found</h1>\n' in response.body.decode()


@pytest.mark.gen_test
def test_init(http_client, base_url):
    'Clear database'
    url = '/api/initialize-Data6ase'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'none'
    assert (yield from assert_success(url, params={'action': 'clear'}))[
        'message'
    ] == 'done'
    assert (yield from assert_success(url, params={'action': 'init'}))[
        'message'
    ] == 'done'
    yield from assert_success(url, params={'action': 'dump'})
    yield from assert_fail(
        url,
        params={'action': 'walk'},
        msg='action should be clear, init, or dump',
    )
    response = yield from server_communication(
        url, params={'action': 'dump', 'human-readable': 'true'}
    )
    assert response.code == 200
    assert 'masonry.min.js' in response.body.decode()


@pytest.mark.gen_test
def test_list_courses(http_client, base_url):
    'Test GET /api/courses'
    url = '/api/courses'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    assert (yield from assert_success(url))['courses'] == ['course1']
    user = 'abigail'
    assert (yield from assert_success(url))['courses'] == ['course2']
    user = 'lawrence'
    assert (yield from assert_success(url))['courses'] == ['course1']
    user = 'eric'
    assert (yield from assert_success(url))['courses'] == ['course2']
    user = 'root'
    assert (yield from assert_success(url))['courses'] == ['course1', 'course2']


@pytest.mark.gen_test
def test_add_course(http_client, base_url):
    'Test POST /api/course/<course_id>'
    url = '/api/course/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'eric'
    yield from assert_fail(
        url + 'course3', method='POST', msg='Permission denied (not admin)'
    )
    user = 'root'
    yield from assert_success(url + 'course3', method='POST')
    assert (yield from assert_success('/api/instructors/course3'))[
        'instructors'
    ] == []
    yield from assert_success(url + 'course3', method='DELETE')
    yield from assert_fail(
        url + 'course3',
        params={'instructors': '"root"]'},
        method='POST',
        msg='Instructors cannot be JSON decoded',
    )
    yield from assert_success(
        url + 'course3', params={'instructors': '["root"]'}, method='POST'
    )
    assert (yield from assert_success('/api/instructors/course3'))[
        'instructors'
    ] == [
        {
            'username': 'root',
            'first_name': None,
            'last_name': None,
            'email': None,
        }
    ]
    yield from assert_fail(
        url + 'course3', method='POST', msg='Course already exists'
    )
    # change owner to eric
    yield from assert_success(
        '/api/instructor/course3/eric',
        method='POST',
        data={'first_name': '', 'last_name': '', 'email': ''},
    )
    assert (yield from assert_success('/api/courses'))['courses'] == [
        'course1',
        'course2',
        'course3',
    ]
    yield from assert_success('/api/instructor/course3/root', method='DELETE')
    user = 'eric'
    assert (yield from assert_success('/api/courses'))['courses'] == [
        'course2',
        'course3',
    ]


@pytest.mark.gen_test
def test_add_instructor(http_client, base_url):
    'Test POST /api/instructor/<course_id>/<instructor_id>'
    url = '/api/instructor/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'eric'
    yield from assert_fail(
        url + 'course2/lawrence',
        method='POST',
        msg='Permission denied (not course instructor)',
    )
    user = 'root'
    yield from assert_fail(
        url + 'course9/lawrence', method='POST', msg='Course not found'
    )
    data = {}
    yield from assert_fail(
        url + 'course2/lawrence',
        data=data,
        method='POST',
        msg='Please supply first name',
    )
    data['first_name'] = 'lawrence_course2_first_name'
    yield from assert_fail(
        url + 'course2/lawrence',
        data=data,
        method='POST',
        msg='Please supply last name',
    )
    data['last_name'] = 'lawrence_course2_last_name'
    yield from assert_fail(
        url + 'course2/lawrence',
        data=data,
        method='POST',
        msg='Please supply email',
    )
    data['email'] = 'lawrence_course2_email'
    yield from assert_success(
        url + 'course2/lawrence', data=data, method='POST'
    )
    assert (
        len(
            (yield from assert_success('/api/instructors/course2'))[
                'instructors'
            ]
        )
        == 2
    )
    # Test changing instructor name
    user = 'abigail'
    yield from assert_fail(
        url + 'course2/lawrence',
        data=data,
        method='POST',
        msg='Permission denied (cannot modify other instructors)',
    )
    user = 'abigail'
    yield from assert_fail(
        url + 'course2/eric',
        data=data,
        method='POST',
        msg='Permission denied (cannot modify instructors)',
    )
    user = 'abigail'
    yield from assert_fail(
        url + 'course2/kevin',
        data=data,
        method='POST',
        msg='Permission denied (cannot modify instructors)',
    )
    user = 'lawrence'
    yield from assert_success(
        url + 'course2/lawrence', data=data, method='POST'
    )
    # Test updating student to instructor, and empty email
    data = {
        'first_name': 'lawrence_course1_first_name',
        'last_name': 'lawrence_course1_last_name',
        'email': '',
    }
    yield from assert_fail(
        url + 'course1/lawrence',
        data=data,
        method='POST',
        msg='Permission denied (not course instructor)',
    )
    user = 'root'
    yield from assert_success(
        url + 'course1/lawrence', data=data, method='POST'
    )
    assert (
        len(
            (yield from assert_success('/api/instructors/course1'))[
                'instructors'
            ]
        )
        == 2
    )
    assert (
        len((yield from assert_success('/api/students/course1'))['students'])
        == 0
    )
    # Test adding non-existing instructor
    data = {'first_name': '', 'last_name': '', 'email': ''}
    yield from assert_success(
        url + 'course3/instructor', data=data, method='POST'
    )


@pytest.mark.gen_test
def test_get_instructor(http_client, base_url):
    'Test GET /api/instructor/<course_id>/<instructor_id>'
    url = '/api/instructor/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'course9/lawrence', msg='Course not found')
    yield from assert_fail(
        url + 'course2/lawrence',
        msg='Permission denied (not related to course)',
    )
    user = 'eric'
    yield from assert_fail(url + 'course9/lawrence', msg='Course not found')
    resp1 = yield from assert_success(url + 'course2/lawrence')
    user = 'abigail'
    yield from assert_fail(url + 'course2/eric', msg='Instructor not found')
    resp2 = yield from assert_success(url + 'course2/lawrence')
    assert resp1 == resp2
    assert resp1['username'] == 'lawrence'
    assert resp1['first_name'] == 'lawrence_course2_first_name'
    assert resp1['last_name'] == 'lawrence_course2_last_name'
    assert resp1['email'] == 'lawrence_course2_email'
    user = 'lawrence'
    resp3 = yield from assert_success(url + 'course1/lawrence')
    assert resp3['username'] == 'lawrence'
    assert resp3['first_name'] == 'lawrence_course1_first_name'
    assert resp3['last_name'] == 'lawrence_course1_last_name'
    assert resp3['email'] == ''


@pytest.mark.gen_test
def test_delete_instructor(http_client, base_url):
    'Test DELETE /api/instructor/<course_id>/<instructor_id>'
    url = '/api/instructor/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'abigail'
    yield from assert_fail(
        url + 'course2/lawrence',
        method='DELETE',
        msg='Permission denied (not admin)',
    )
    user = 'root'
    yield from assert_fail(
        url + 'course9/lawrence', method='DELETE', msg='Course not found'
    )
    yield from assert_fail(
        url + 'course2/eric', method='DELETE', msg='Instructor not found'
    )
    yield from assert_success(url + 'course2/lawrence', method='DELETE')
    yield from assert_success(url + 'course1/kevin', method='DELETE')
    yield from assert_success(
        url + 'course1/kevin',
        method='POST',
        data={'first_name': '', 'last_name': '', 'email': ''},
    )


@pytest.mark.gen_test
def test_list_instructors(http_client, base_url):
    'Test GET /api/instructors/<course_id>'
    url = '/api/instructors/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'course9', msg='Course not found')
    yield from assert_fail(
        url + 'course2', msg='Permission denied (not related to course)'
    )
    user = 'eric'
    resp1 = (yield from assert_success(url + 'course2'))['instructors']
    user = 'abigail'
    resp2 = (yield from assert_success(url + 'course2'))['instructors']
    assert resp1 == resp2
    assert len(resp1) == 1
    assert resp1[0]['username'] == 'abigail'
    assert resp1[0]['first_name'] is None
    assert resp1[0]['last_name'] is None
    assert resp1[0]['email'] is None


@pytest.mark.gen_test
def test_add_student(http_client, base_url):
    'Test POST /api/student/<course_id>/<student_id>'
    url = '/api/student/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'eric'
    yield from assert_fail(
        url + 'course9/lawrence', method='POST', msg='Course not found'
    )
    yield from assert_fail(
        url + 'course2/lawrence',
        method='POST',
        msg='Permission denied (not course instructor)',
    )
    user = 'abigail'
    data = {}
    yield from assert_fail(
        url + 'course2/lawrence',
        data=data,
        method='POST',
        msg='Please supply first name',
    )
    data['first_name'] = 'lawrence_course2_first_name'
    yield from assert_fail(
        url + 'course2/lawrence',
        data=data,
        method='POST',
        msg='Please supply last name',
    )
    data['last_name'] = 'lawrence_course2_last_name'
    yield from assert_fail(
        url + 'course2/lawrence',
        data=data,
        method='POST',
        msg='Please supply email',
    )
    data['email'] = 'lawrence_course2_email'
    yield from assert_success(
        url + 'course2/lawrence', data=data, method='POST'
    )
    assert (
        len((yield from assert_success('/api/students/course2'))['students'])
        == 2
    )
    # Test updating instructor to student, and empty email
    yield from assert_fail(
        url + 'course2/abigail',
        data=data,
        method='POST',
        msg='Cannot add instructor as student',
    )
    user = 'kevin'
    data = {
        'first_name': 'lawrence_course1_first_name',
        'last_name': 'lawrence_course1_last_name',
        'email': '',
    }
    yield from assert_fail(
        url + 'course1/lawrence',
        data=data,
        method='POST',
        msg='Cannot add instructor as student',
    )
    assert (
        len(
            (yield from assert_success('/api/instructors/course1'))[
                'instructors'
            ]
        )
        == 2
    )
    assert (
        len((yield from assert_success('/api/students/course1'))['students'])
        == 0
    )
    user = 'root'
    yield from assert_success(
        '/api/instructor/course1/lawrence', method='DELETE'
    )
    user = 'kevin'
    yield from assert_success(
        url + 'course1/lawrence', data=data, method='POST'
    )
    assert (
        len(
            (yield from assert_success('/api/instructors/course1'))[
                'instructors'
            ]
        )
        == 1
    )
    assert (
        len((yield from assert_success('/api/students/course1'))['students'])
        == 1
    )
    # Test adding non-existing instructor
    user = 'eric'
    data = {'first_name': '', 'last_name': '', 'email': ''}
    yield from assert_success(url + 'course3/student', data=data, method='POST')


@pytest.mark.gen_test
def test_get_student(http_client, base_url):
    'Test GET /api/student/<course_id>/<student_id>'
    url = '/api/student/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'course9/lawrence', msg='Course not found')
    yield from assert_fail(
        url + 'course2/lawrence',
        msg='Permission denied (not course instructor)',
    )
    user = 'eric'
    yield from assert_fail(
        url + 'course2/lawrence',
        msg='Permission denied (not course instructor)',
    )
    user = 'abigail'
    yield from assert_fail(url + 'course2/abigail', msg='Student not found')
    resp = yield from assert_success(url + 'course2/lawrence')
    assert resp['username'] == 'lawrence'
    assert resp['first_name'] == 'lawrence_course2_first_name'
    assert resp['last_name'] == 'lawrence_course2_last_name'
    assert resp['email'] == 'lawrence_course2_email'


@pytest.mark.gen_test
def test_delete_student(http_client, base_url):
    'Test DELETE /api/student/<course_id>/<student_id>'
    url = '/api/student/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'eric'
    yield from assert_fail(
        url + 'course9/lawrence', method='DELETE', msg='Course not found'
    )
    yield from assert_fail(
        url + 'course2/lawrence',
        method='DELETE',
        msg='Permission denied (not course instructor)',
    )
    user = 'abigail'
    yield from assert_fail(
        url + 'course2/kevin', method='DELETE', msg='Student not found'
    )
    yield from assert_success(url + 'course2/lawrence', method='DELETE')


@pytest.mark.gen_test
def test_add_students(http_client, base_url):
    'Test POST /api/students/<course_id>'
    url = '/api/students/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'course9', msg='Course not found')
    yield from assert_fail(
        url + 'course2',
        method='POST',
        msg='Permission denied (not course instructor)',
    )
    yield from assert_fail(
        url + 'course1', method='POST', data={}, msg='Please supply students'
    )
    yield from assert_fail(
        url + 'course1',
        method='POST',
        data={'students': '"'},
        msg='Students cannot be JSON decoded',
    )
    yield from assert_fail(
        url + 'course1',
        method='POST',
        data={'students': '12'},
        msg='Incorrect request format',
    )
    yield from assert_fail(
        url + 'course1',
        method='POST',
        data={'students': '[]'},
        msg='Please supply students',
    )
    yield from assert_fail(
        url + 'course1',
        method='POST',
        data={'students': '[1,2]'},
        msg='Incorrect request format',
    )
    students = [{'username': 'a', 'email': 'b', 'first_name': 'c'}]
    yield from assert_fail(
        url + 'course1',
        method='POST',
        data={'students': json.dumps(students)},
        msg='Incorrect request format',
    )
    students = [
        {'username': 'a', 'first_name': 'af', 'last_name': 'al', 'email': 'ae'},
        {'username': 'b', 'first_name': 'bf', 'last_name': 'bl', 'email': 'be'},
        {'username': 'c', 'first_name': 'cf', 'last_name': 'cl', 'email': 'ce'},
        {'username': 'd', 'first_name': 'df', 'last_name': 'dl', 'email': 'de'},
        {'username': 'e', 'first_name': '', 'last_name': '', 'email': ''},
        {'username': 'kevin', 'first_name': '', 'last_name': '', 'email': ''},
        {
            'username': 'lawrence',
            'first_name': '',
            'last_name': '',
            'email': '',
        },
    ]
    resp = yield from assert_success(
        url + 'course1', method='POST', data={'students': json.dumps(students)}
    )
    expected = [
        {'username': 'a', 'success': True},
        {'username': 'b', 'success': True},
        {'username': 'c', 'success': True},
        {'username': 'd', 'success': True},
        {'username': 'e', 'success': True},
        {
            'username': 'kevin',
            'success': False,
            'message': 'Cannot add instructor as student',
        },
        {'username': 'lawrence', 'success': True},
    ]
    assert resp['status'] == expected
    resp = (yield from assert_success(url + 'course1'))['students']
    assert len(resp) == 6
    for i in resp:
        assert i in students


@pytest.mark.gen_test
def test_list_students(http_client, base_url):
    'Test GET /api/students/<course_id>'
    url = '/api/students/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'course9', msg='Course not found')
    yield from assert_fail(
        url + 'course2', msg='Permission denied (not course instructor)'
    )
    user = 'eric'
    yield from assert_fail(
        url + 'course2', msg='Permission denied (not course instructor)'
    )
    user = 'abigail'
    resp = (yield from assert_success(url + 'course2'))['students']
    assert len(resp) == 1
    assert resp[0]['username'] == 'eric'
    assert resp[0]['first_name'] is None
    assert resp[0]['last_name'] is None
    assert resp[0]['email'] is None


@pytest.mark.gen_test
def test_list_assignments(http_client, base_url):
    'Test GET /api/assignments/<course_id>'
    url = '/api/assignments/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(
        url + 'course2', msg='Permission denied (not related to course)'
    )
    user = 'abigail'
    assert (yield from assert_success(url + 'course2'))['assignments'] == [
        'assignment2a',
        'assignment2b',
    ]
    user = 'lawrence'
    yield from assert_fail(
        url + 'course2', msg='Permission denied (not related to course)'
    )
    user = 'eric'
    assert (yield from assert_success(url + 'course2'))['assignments'] == [
        'assignment2a',
        'assignment2b',
    ]
    yield from assert_fail(url + 'jkl', msg='Course not found')


@pytest.mark.gen_test
def test_download_assignment(http_client, base_url):
    'Test GET /api/assignment/<course_id>/<assignment_id>'
    url = '/api/assignment/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    files = (yield from assert_success(url + 'course1/challenge'))['files']
    assert files[0]['path'] == 'file2'
    assert base64.b64decode(files[0]['content'].encode()) == b'22222'
    assert files[0]['checksum'] == hashlib.md5(b'22222').hexdigest()
    yield from assert_fail(url + 'jkl/challenger', msg='Course not found')
    yield from assert_fail(
        url + 'course1/challenger', msg='Assignment not found'
    )
    # Check list_only
    files = (
        yield from assert_success(url + 'course1/challenge?list_only=true')
    )['files']
    assert set(files[0]) == {'path', 'checksum'}
    assert files[0]['path'] == 'file2'
    assert files[0]['checksum'] == hashlib.md5(b'22222').hexdigest()
    user = 'eric'
    yield from assert_fail(
        url + 'course1/challenge',
        msg='Permission denied (not related to course)',
    )


@pytest.mark.gen_test
def test_release_assignment(http_client, base_url):
    'Test POST /api/assignment/<course_id>/<assignment_id>'
    url = '/api/assignment/'
    global hc, bu, user
    hc, bu = http_client, base_url
    data = {
        'files': json.dumps(
            [
                {'path': 'a', 'content': 'amtsCg=='},
                {'path': 'b', 'content': 'amtsCg=='},
            ]
        )
    }
    user = 'kevin'
    yield from assert_fail(
        url + 'jkl/challenger', method='POST', data=data, msg='Course not found'
    )
    yield from assert_fail(
        url + 'course1/challenger', method='POST', msg='Please supply files'
    )
    yield from assert_success(
        url + 'course1/challenger', method='POST', data=data
    )
    yield from assert_fail(
        url + 'course1/challenger',
        method='POST',
        data=data,
        msg='Assignment already exists',
    )
    data['files'] = json.dumps([{'path': 'a', 'content': 'amtsCg'}])
    yield from assert_fail(
        url + 'course1/challenges',
        method='POST',
        data=data,
        msg='Content cannot be base64 decoded',
    )
    for pathname in ['/a', '/', '', '../etc', 'a/./a.py', 'a/.']:
        data['files'] = json.dumps([{'path': pathname, 'content': ''}])
        yield from assert_fail(
            url + 'course1/challenges',
            method='POST',
            data=data,
            msg='Illegal path',
        )
    user = 'abigail'
    yield from assert_fail(
        url + 'course1/challenger',
        method='POST',
        data=data,
        msg='Permission denied (not course instructor)',
    )
    user = 'lawrence'
    yield from assert_fail(
        url + 'course1/challenger',
        method='POST',
        data=data,
        msg='Permission denied (not course instructor)',
    )
    user = 'eric'
    yield from assert_fail(
        url + 'course1/challenger',
        method='POST',
        data=data,
        msg='Permission denied (not course instructor)',
    )


@pytest.mark.gen_test
def test_delete_assignment(http_client, base_url):
    'Test DELETE /api/assignment/<course_id>/<assignment_id>'
    url = '/api/assignment/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'lawrence'
    yield from assert_fail(
        url + 'course1/challenger',
        method='DELETE',
        msg='Permission denied (not course instructor)',
    )
    user = 'kevin'
    yield from assert_fail(
        url + 'jkl/challenger', method='DELETE', msg='Course not found'
    )
    yield from assert_fail(
        url + 'course1/challengers', method='DELETE', msg='Assignment not found'
    )
    yield from assert_success(url + 'course1/challenger')
    yield from assert_success(url + 'course1/challenger', method='DELETE')
    yield from assert_fail(
        url + 'course1/challenger', msg='Assignment not found'
    )


@pytest.mark.gen_test
def test_list_submissions(http_client, base_url):
    'Test GET /api/submissions/<course_id>/<assignment_id>'
    url = '/api/submissions/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'jkl/challenge', msg='Course not found')
    yield from assert_fail(
        url + 'course1/challenges', msg='Assignment not found'
    )
    result = yield from assert_success(url + 'course1/challenge')
    assert len(result['submissions']) == 2
    assert set(result['submissions'][0]) == {'student_id', 'timestamp'}
    assert result['submissions'][0]['student_id'] == 'lawrence'
    assert result['submissions'][1]['student_id'] == 'lawrence'
    user = 'abigail'
    result = yield from assert_success(url + 'course2/assignment2a')
    assert len(result['submissions']) == 0
    user = 'eric'
    yield from assert_fail(
        url + 'course1/challenges',
        msg='Permission denied (not course instructor)',
    )
    yield from assert_fail(
        url + 'course2/assignment2a',
        msg='Permission denied (not course instructor)',
    )


@pytest.mark.gen_test
def test_list_student_submission(http_client, base_url):
    'Test GET /api/submissions/<course_id>/<assignment_id>/<student_id>'
    url = '/api/submissions/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'jkl/challenge/st', msg='Course not found')
    yield from assert_fail(
        url + 'course1/challenges/st', msg='Assignment not found'
    )
    yield from assert_fail(
        url + 'course1/challenge/st', msg='Student not found'
    )
    result = yield from assert_success(url + 'course1/challenge/lawrence')
    assert len(result['submissions']) == 2
    assert set(result['submissions'][0]) == {'student_id', 'timestamp'}
    user = 'eric'
    result = yield from assert_success(url + 'course2/assignment2a/eric')
    assert len(result['submissions']) == 0
    user = 'kevin'
    yield from assert_fail(
        url + 'course2/assignment2a/eric',
        msg='Permission denied (not course instructor)',
    )
    user = 'abigail'
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        msg='Permission denied (not course instructor)',
    )
    user = 'lawrence'
    yield from assert_success(url + 'course1/challenge/lawrence')
    user = 'eric'
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        msg='Permission denied (not course instructor)',
    )


@pytest.mark.gen_test
def test_submit_assignment(http_client, base_url):
    'Test POST /api/submission/<course_id>/<assignment_id>'
    url = '/api/submission/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    data = {
        'files': json.dumps(
            [
                {'path': 'a', 'content': 'amtsCg=='},
                {'path': 'b', 'content': 'amtsCg=='},
            ]
        )
    }
    yield from assert_fail(
        url + 'jkl/challenge', method='POST', msg='Course not found'
    )
    yield from assert_fail(
        url + 'course1/challenges', method='POST', msg='Assignment not found'
    )
    user = 'lawrence'
    yield from assert_fail(
        url + 'course1/challenge', method='POST', msg='Please supply files'
    )
    resp1 = yield from assert_success(
        url + 'course1/challenge', method='POST', data=data
    )
    ts1 = MyHelpers().strptime(resp1['timestamp'])
    data['files'] = json.dumps([{'path': 'a', 'content': 'amtsCg=='}])
    resp2 = yield from assert_success(
        url + 'course1/challenge', method='POST', data=data
    )
    ts2 = MyHelpers().strptime(resp2['timestamp'])
    assert ts1 < ts2
    assert ts2 < ts1 + datetime.timedelta(seconds=1)
    data['files'] = json.dumps([{'path': 'a', 'content': 'amtsCg'}])
    yield from assert_fail(
        url + 'course1/challenge',
        method='POST',
        data=data,
        msg='Content cannot be base64 decoded',
    )
    data['files'] = 'a-random-string'
    yield from assert_fail(
        url + 'course1/challenge',
        method='POST',
        data=data,
        msg='Files cannot be JSON decoded',
    )
    user = 'kevin'
    result = yield from assert_success('/api/submissions/course1/challenge')
    assert len(result['submissions']) == 4  # 2 from init, 2 from this
    user = 'eric'
    yield from assert_fail(
        url + 'course1/challenge',
        method='POST',
        msg='Permission denied (not related to course)',
    )


@pytest.mark.gen_test
def test_download_submission(http_client, base_url):
    'Test GET /api/submission/<course_id>/<assignment_id>/<student_id>'
    url = '/api/submission/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'jkl/challenge/st', msg='Course not found')
    yield from assert_fail(
        url + 'course1/challenges/st', msg='Assignment not found'
    )
    yield from assert_fail(
        url + 'course1/challenge/st', msg='Student not found'
    )
    # Test get latest
    result = yield from assert_success(url + 'course1/challenge/lawrence')
    files = result['files']
    assert len(files) == 1
    file_obj = next(filter(lambda x: x['path'] == 'a', files), None)
    assert base64.b64decode(file_obj['content'].encode()) == b'jkl\n'
    assert file_obj['checksum'] == hashlib.md5(b'jkl\n').hexdigest()
    user = 'abigail'
    yield from assert_fail(
        url + 'course2/assignment2a/eric', msg='Submission not found'
    )
    # Test get latest with list_only
    user = 'kevin'
    result = yield from assert_success(
        url + 'course1/challenge/lawrence', params={'list_only': 'true'}
    )
    files = result['files']
    assert len(files) == 1
    assert set(files[0]) == {'path', 'checksum'}
    assert files[0]['path'] == 'a'
    assert files[0]['checksum'] == hashlib.md5(b'jkl\n').hexdigest()
    # Test timestamp
    result = yield from assert_success(
        '/api/submissions/course1/challenge/lawrence'
    )
    timestamp = sorted(map(lambda x: x['timestamp'], result['submissions']))[0]
    result = yield from assert_success(
        url + 'course1/challenge/lawrence', params={'timestamp': timestamp}
    )
    files = result['files']
    assert len(files) == 1
    file_obj = next(filter(lambda x: x['path'] == 'file3', files), None)
    assert base64.b64decode(file_obj['content'].encode()) == b'33333'
    assert file_obj['checksum'] == hashlib.md5(b'33333').hexdigest()
    # Test timestamp with list_only
    result = yield from assert_success(
        '/api/submissions/course1/challenge/lawrence'
    )
    timestamp = sorted(map(lambda x: x['timestamp'], result['submissions']))[0]
    result = yield from assert_success(
        url + 'course1/challenge/lawrence',
        params={'timestamp': timestamp, 'list_only': 'true'},
    )
    files = result['files']
    assert len(files) == 1
    file_obj = next(filter(lambda x: x['path'] == 'file3', files), None)
    assert 'content' not in file_obj
    assert file_obj['checksum'] == hashlib.md5(b'33333').hexdigest()
    # Test timestamp not found
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f %Z')
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        params={'timestamp': timestamp},
        msg='Submission not found',
    )
    # Test permission
    user = 'eric'
    yield from assert_fail(
        url + 'course2/assignment2a/eric',
        msg='Permission denied (not course instructor)',
    )


@pytest.mark.gen_test
def test_upload_feedback(http_client, base_url):
    'Test POST /api/feedback/<course_id>/<assignment_id>/<student_id>'
    url = '/api/feedback/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    data = {
        'files': json.dumps(
            [
                {'path': 'a', 'content': 'amtsCg=='},
                {'path': 'b', 'content': 'amtsCg=='},
            ]
        ),
        'timestamp': '2020-01-01 00:00:00.000000 ',
    }
    yield from assert_fail(
        url + 'jkl/challenge/st',
        method='POST',
        data=data,
        msg='Course not found',
    )
    yield from assert_fail(
        url + 'course1/challenges/st',
        method='POST',
        data=data,
        msg='Assignment not found',
    )
    yield from assert_fail(
        url + 'course1/challenge/st',
        method='POST',
        data=data,
        msg='Student not found',
    )
    yield from assert_success(
        url + 'course1/challenge/lawrence', method='POST', data=data
    )
    data['files'] = json.dumps([{'path': 'c', 'content': 'amtsCf=='}])
    yield from assert_success(
        url + 'course1/challenge/lawrence', method='POST', data=data
    )
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        method='POST',
        data={},
        msg='Please supply timestamp',
    )
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        method='POST',
        data={'timestamp': 'a'},
        msg='Time format incorrect',
    )
    user = 'abigail'
    yield from assert_fail(
        url + 'course2/assignment2a/eric',
        method='POST',
        data=data,
        msg='Submission not found',
    )
    yield from assert_fail(
        url + 'course2/assignment2a/eric',
        method='POST',
        data={'timestamp': data['timestamp']},
        msg='Submission not found',
    )
    user = 'eric'
    yield from assert_fail(
        url + 'course2/assignment2a/eric',
        method='POST',
        data=data,
        msg='Permission denied (not course instructor)',
    )


@pytest.mark.gen_test
def test_download_feedback(http_client, base_url):
    'Test GET /api/feedback/<course_id>/<assignment_id>/<student_id>'
    url = '/api/feedback/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(url + 'jkl/challenge/st', msg='Course not found')
    yield from assert_fail(
        url + 'course1/challenges/st', msg='Assignment not found'
    )
    yield from assert_fail(
        url + 'course1/challenge/st', msg='Student not found'
    )
    meta = yield from assert_success(
        '/api/submission/course1/challenge/lawrence'
    )
    timestamp = meta['timestamp']
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        params={},
        msg='Please supply timestamp',
    )
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        params={'timestamp': 'a'},
        msg='Time format incorrect',
    )
    user = 'eric'
    yield from assert_fail(
        url + 'course2/assignment2a/eric',
        params={'timestamp': timestamp},
        msg='Submission not found',
    )
    user = 'kevin'
    feedback = yield from assert_success(
        url + 'course1/challenge/lawrence', params={'timestamp': timestamp}
    )
    assert feedback['files'] == []
    # Submit again ('amtsDg==' is 'jkl\x0e')
    data = {
        'files': json.dumps([{'path': 'a', 'content': 'amtsDg=='}]),
        'timestamp': timestamp,
    }
    yield from assert_success(
        url + 'course1/challenge/lawrence', method='POST', data=data
    )
    # Fetch again
    feedback = yield from assert_success(
        url + 'course1/challenge/lawrence', params={'timestamp': timestamp}
    )
    assert len(feedback['files']) == 1
    assert feedback['files'][0]['path'] == 'a'
    file_obj = feedback['files'][0]
    assert base64.b64decode(file_obj['content'].encode()) == b'jkl\x0e'
    assert file_obj['checksum'] == hashlib.md5(b'jkl\x0e').hexdigest()
    # Again, submit again ('nmtsDg==' is 'nkl\x0e')
    data = {
        'files': json.dumps([{'path': 'a', 'content': 'bmtsDg=='}]),
        'timestamp': timestamp,
    }
    yield from assert_success(
        url + 'course1/challenge/lawrence', method='POST', data=data
    )
    # Again, fetch again
    feedback = yield from assert_success(
        url + 'course1/challenge/lawrence', params={'timestamp': timestamp}
    )
    assert len(feedback['files']) == 1
    file_obj = feedback['files'][0]
    assert file_obj['path'] == 'a'
    assert base64.b64decode(file_obj['content'].encode()) == b'nkl\x0e'
    assert file_obj['checksum'] == hashlib.md5(b'nkl\x0e').hexdigest()
    # Check list_only
    feedback = yield from assert_success(
        url + 'course1/challenge/lawrence',
        params={'timestamp': timestamp, 'list_only': 'true'},
    )
    assert len(feedback['files']) == 1
    assert set(feedback['files'][0]) == {'path', 'checksum'}
    assert file_obj['checksum'] == hashlib.md5(b'nkl\x0e').hexdigest()
    assert feedback['files'][0]['path'] == 'a'
    # Permission check
    user = 'kevin'
    yield from assert_fail(
        url + 'course1/challenge/lawrence', msg='Please supply timestamp'
    )
    user = 'abigail'
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        msg='Permission denied (not course instructor)',
    )
    user = 'lawrence'
    yield from assert_fail(
        url + 'course1/challenge/lawrence', msg='Please supply timestamp'
    )
    user = 'eric'
    yield from assert_fail(
        url + 'course1/challenge/lawrence',
        msg='Permission denied (not course instructor)',
    )


@pytest.mark.gen_test
def test_remove_course(http_client, base_url):
    'Test DELETE /api/course/<course_id>'
    url = '/api/course/'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'kevin'
    yield from assert_fail(
        url + 'course1', method='DELETE', msg='Permission denied (not admin)'
    )
    user = 'root'
    yield from assert_success(url + 'course1', method='DELETE')
    yield from assert_success(url + 'course2', method='DELETE')
    yield from assert_success(url + 'course3', method='DELETE')
    yield from assert_fail(
        url + 'course4', method='DELETE', msg='Course not found'
    )
    resp = yield from assert_success(
        '/api/initialize-Data6ase', params={'action': 'dump'}
    )
    # All other tables should be empty
    assert set(resp.keys()) == {'success', 'users'}


@pytest.mark.gen_test
def test_corner_case(http_client, base_url):
    'Test corner cases to increase coverage'
    global hc, bu, user
    hc, bu = http_client, base_url
    init_url = '/api/initialize-Data6ase'
    user = 'none'
    yield from assert_success(init_url, params={'action': 'clear'})
    yield from assert_success(init_url, params={'action': 'init'})
    user = 'eric'
    data = {
        'files': json.dumps(
            [{'path': 'a.abcdefghijklmnopqrstuvw', 'content': 'amtsCg=='}]
        )
    }
    yield from assert_success(
        '/api/submission/course2/assignment2a', method='POST', data=data,
    )
    user = 'abigail'
    assert (
        yield from assert_success('/api/submission/course2/assignment2a/eric')
    )['files'][0]['path'] == 'a.abcdefghijklmnopqrstuvw'


@pytest.mark.gen_test
def test_nodebug(http_client, base_url):
    'Test ngshare with debug-mode off'
    application.debug = False
    url = '/api/initialize-Data6ase'
    global hc, bu, user
    hc, bu = http_client, base_url
    user = 'none'
    yield from assert_fail(
        url, params={'action': 'clear'}, msg='Debug mode is off'
    )
    yield from assert_fail(
        url, params={'action': 'init'}, msg='Debug mode is off'
    )
    response = yield from server_communication('/api/random-page')
    assert response.code == 404
    assert response.body.decode() == '<h1>404 Not Found</h1>\n'


def test_notimpl():
    'Test NotImplementedError etc'
    try:
        MyHelpers().json_error(404, 'Not Found')
    except NotImplementedError:
        pass
    assert MockAuth().get_login_url().startswith('http')


def test_clean():
    'Clean temporary files'
    global db_name, storage_name
    os.remove(db_name)
    shutil.rmtree(storage_name)



"""Init

Revision ID: aa00db20c10a
Revises: 
Create Date: 2020-04-20 00:34:25.543691

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'aa00db20c10a'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'courses',
        sa.Column('_id', sa.INTEGER(), nullable=False),
        sa.Column('id', sa.TEXT(), nullable=True),
        sa.PrimaryKeyConstraint('_id'),
        sa.UniqueConstraint('id'),
    )
    op.create_table(
        'files',
        sa.Column('_id', sa.INTEGER(), nullable=False),
        sa.Column('filename', sa.TEXT(), nullable=True),
        sa.Column('checksum', sa.TEXT(), nullable=True),
        sa.Column('actual_name', sa.TEXT(), nullable=True),
        sa.PrimaryKeyConstraint('_id'),
    )
    op.create_table(
        'users',
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
    )
    op.create_table(
        'assignments',
        sa.Column('_id', sa.INTEGER(), nullable=False),
        sa.Column('id', sa.TEXT(), nullable=True),
        sa.Column('course_id', sa.INTEGER(), nullable=True),
        sa.Column('due', sa.TIMESTAMP(), nullable=True),
        sa.ForeignKeyConstraint(['course_id'], ['courses._id'],),
        sa.PrimaryKeyConstraint('_id'),
    )
    op.create_table(
        'instructor_assoc_table',
        sa.Column('left_id', sa.TEXT(), nullable=False),
        sa.Column('right_id', sa.TEXT(), nullable=False),
        sa.Column('first_name', sa.TEXT(), nullable=True),
        sa.Column('last_name', sa.TEXT(), nullable=True),
        sa.Column('email', sa.TEXT(), nullable=True),
        sa.ForeignKeyConstraint(['left_id'], ['users.id'],),
        sa.ForeignKeyConstraint(['right_id'], ['courses.id'],),
        sa.PrimaryKeyConstraint('left_id', 'right_id'),
    )
    op.create_table(
        'student_assoc_table',
        sa.Column('left_id', sa.TEXT(), nullable=False),
        sa.Column('right_id', sa.TEXT(), nullable=False),
        sa.Column('first_name', sa.TEXT(), nullable=True),
        sa.Column('last_name', sa.TEXT(), nullable=True),
        sa.Column('email', sa.TEXT(), nullable=True),
        sa.ForeignKeyConstraint(['left_id'], ['users.id'],),
        sa.ForeignKeyConstraint(['right_id'], ['courses.id'],),
        sa.PrimaryKeyConstraint('left_id', 'right_id'),
    )
    op.create_table(
        'assignment_files_assoc_table',
        sa.Column('left_id', sa.TEXT(), nullable=False),
        sa.Column('right_id', sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(['left_id'], ['assignments._id'],),
        sa.ForeignKeyConstraint(['right_id'], ['files._id'],),
        sa.PrimaryKeyConstraint('left_id', 'right_id'),
    )
    op.create_table(
        'submissions',
        sa.Column('_id', sa.INTEGER(), nullable=False),
        sa.Column('assignment_id', sa.INTEGER(), nullable=True),
        sa.Column('timestamp', sa.TIMESTAMP(), nullable=True),
        sa.Column('student_id', sa.TEXT(), nullable=True),
        sa.ForeignKeyConstraint(['assignment_id'], ['assignments._id'],),
        sa.ForeignKeyConstraint(['student_id'], ['users.id'],),
        sa.PrimaryKeyConstraint('_id'),
    )
    op.create_table(
        'feedback_files_assoc_table',
        sa.Column('left_id', sa.TEXT(), nullable=False),
        sa.Column('right_id', sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(['left_id'], ['submissions._id'],),
        sa.ForeignKeyConstraint(['right_id'], ['files._id'],),
        sa.PrimaryKeyConstraint('left_id', 'right_id'),
    )
    op.create_table(
        'submission_files_assoc_table',
        sa.Column('left_id', sa.TEXT(), nullable=False),
        sa.Column('right_id', sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(['left_id'], ['submissions._id'],),
        sa.ForeignKeyConstraint(['right_id'], ['files._id'],),
        sa.PrimaryKeyConstraint('left_id', 'right_id'),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('submission_files_assoc_table')
    op.drop_table('feedback_files_assoc_table')
    op.drop_table('submissions')
    op.drop_table('assignment_files_assoc_table')
    op.drop_table('student_assoc_table')
    op.drop_table('instructor_assoc_table')
    op.drop_table('assignments')
    op.drop_table('users')
    op.drop_table('files')
    op.drop_table('courses')
    # ### end Alembic commands ###



"""Add file size

Revision ID: 1921a169739b
Revises: aa00db20c10a
Create Date: 2020-04-30 14:39:41.662130

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1921a169739b'
down_revision = 'aa00db20c10a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('files', sa.Column('size', sa.INTEGER(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # The following command won't work, see:
    # https://alembic.sqlalchemy.org/en/latest/batch.html
    # op.drop_column('files', 'size')
    # Therefore, we have to do:
    with op.batch_alter_table("files") as batch_op:
        batch_op.drop_column('size')

    # ### end Alembic commands ###



'''
    Database structure for NGShare
'''

import datetime
import hashlib

from sqlalchemy import (
    Table,
    Column,
    INTEGER,
    TEXT,
    TIMESTAMP,
    BOOLEAN,
    ForeignKey,
)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

# Assignment -> Course (One to Many)
assignment_files_assoc_table = Table(
    'assignment_files_assoc_table',
    Base.metadata,
    Column('left_id', TEXT, ForeignKey('assignments._id'), primary_key=True),
    Column('right_id', INTEGER, ForeignKey('files._id'), primary_key=True),
)

# Submission -> Course (One to Many)
submission_files_assoc_table = Table(
    'submission_files_assoc_table',
    Base.metadata,
    Column('left_id', TEXT, ForeignKey('submissions._id'), primary_key=True),
    Column('right_id', INTEGER, ForeignKey('files._id'), primary_key=True),
)

# Submission (feedback) -> Course (One to Many)
feedback_files_assoc_table = Table(
    'feedback_files_assoc_table',
    Base.metadata,
    Column('left_id', TEXT, ForeignKey('submissions._id'), primary_key=True),
    Column('right_id', INTEGER, ForeignKey('files._id'), primary_key=True),
)


class User(Base):
    'A JupyterHub user; can be either instructor or student, or both'
    __tablename__ = 'users'
    id = Column(TEXT, primary_key=True)
    teaching = association_proxy(
        'inst_assoc',
        'course',
        creator=lambda course: InstructorAssociation(course=course),
        cascade_scalar_deletes=True,
    )
    taking = association_proxy(
        'student_assoc',
        'course',
        creator=lambda course: StudentAssociation(course=course),
        cascade_scalar_deletes=True,
    )

    def __init__(self, name):
        'Initialize with JupyterHub user name'
        self.id = name

    def __str__(self):
        return '<User %s>' % self.id

    def dump(self):
        'Dump data to dict'
        return {
            'id': self.id,
        }

    def delete(self, db):
        'Remove user and dependent data'
        raise NotImplementedError('Currently users cannot be deleted')

    @staticmethod
    def from_jupyterhub_user(user_model, db):
        'Import users from JupyterHub'
        user_name = user_model['name']
        user = db.query(User).filter(User.id == user_name).one_or_none()
        if user is None:
            user = User(user_name)
            db.add(user)
            db.commit()
        return user


class Course(Base):
    'An nbgrader course'
    __tablename__ = 'courses'
    # in case course name needs to be changed
    _id = Column(INTEGER, primary_key=True)
    id = Column(TEXT, unique=True)
    instructors = association_proxy(
        'inst_assoc',
        'user',
        creator=lambda user: InstructorAssociation(user=user,),
        cascade_scalar_deletes=True,
    )
    students = association_proxy(
        'student_assoc',
        'user',
        creator=lambda user: StudentAssociation(user=user,),
        cascade_scalar_deletes=True,
    )
    assignments = relationship('Assignment', backref='course')

    def dump(self):
        'Dump data to dict'
        return {
            '_id': self._id,
            'id': self.id,
        }

    def __init__(self, name, instructors=()):
        'Initialize with course name and teacher'
        self.id = name
        for instructor in instructors:
            self.instructors.append(instructor)

    def __str__(self):
        return '<Course %s>' % self.id

    def delete(self, db):
        'Remove course and dependent data'
        for assignment in self.assignments:
            assignment.delete(db)
        for student in self.students[:]:
            self.students.remove(student)
        for instructor in self.instructors[:]:
            self.instructors.remove(instructor)
        db.delete(self)


class Assignment(Base):
    'An nbgrader assignment'
    __tablename__ = 'assignments'
    # in case assignment name needs to be changed
    _id = Column(INTEGER, primary_key=True)
    id = Column(TEXT)
    course_id = Column(INTEGER, ForeignKey('courses._id'))
    submissions = relationship('Submission', backref='assignment')
    files = relationship('File', secondary=assignment_files_assoc_table)
    released = BOOLEAN()
    due = Column(TIMESTAMP)
    # TODO: timezoon

    def __init__(self, name, course):
        'Initialize with assignment name and course'
        self.id = name
        self.course = course

    def __str__(self):
        return '<Assignment %s>' % self.id

    def dump(self):
        'Dump data to dict'
        return {
            '_id': self._id,
            'id': self.id,
            'course_id': self.course_id,
            'released': bool(self.released),
            'due': self.due and self.due.strftime('%Y-%m-%d %H:%M:%S.%f %Z'),
        }

    def delete(self, db):
        'Remove assignment and dependent data (files, submissions)'
        for file_obj in self.files[:]:
            file_obj.delete(db)
            self.files.remove(file_obj)
        for submission in self.submissions[:]:
            submission.delete(db)
            self.submissions.remove(submission)
        db.delete(self)


class Submission(Base):
    'A submission for an assignment'
    __tablename__ = 'submissions'
    _id = Column(INTEGER, primary_key=True)
    assignment_id = Column(INTEGER, ForeignKey('assignments._id'))
    timestamp = Column(TIMESTAMP)
    student_id = Column(TEXT, ForeignKey('users.id'))
    files = relationship('File', secondary=submission_files_assoc_table)
    feedbacks = relationship('File', secondary=feedback_files_assoc_table)
    student = relationship('User')

    def __init__(self, student, assignment):
        'Initialize with student and assignment'
        self.timestamp = datetime.datetime.now(datetime.timezone.utc)
        self.student = student
        self.assignment = assignment

    def __str__(self):
        return '<Submission %d>' % self._id

    def dump(self):
        'Dump data to dict'
        return {
            '_id': self._id,
            'assignment_id': self.assignment_id,
            'timestamp': self.timestamp
            and self.timestamp.strftime('%Y-%m-%d %H:%M:%S.%f %Z'),
            'student_id': self.student_id,
        }

    def delete(self, db):
        'Remove submission and dependent data (files, feedbacks)'
        for file_obj in self.files[:]:
            file_obj.delete(db)
            self.files.remove(file_obj)
        for file_obj in self.feedbacks[:]:
            file_obj.delete(db)
            self.feedbacks.remove(file_obj)
        db.delete(self)


class File(Base):
    'A File (for assignment, submission file, or submission feedback)'
    __tablename__ = 'files'
    _id = Column(INTEGER, primary_key=True)
    filename = Column(TEXT)
    checksum = Column(TEXT)
    size = Column(INTEGER)
    actual_name = Column(TEXT)

    def __init__(self, filename, contents, actual_name=None):
        'Initialize with file name and content; auto-compute md5 and size'
        self.filename = filename
        self.checksum = hashlib.md5(contents).hexdigest()
        self.size = len(contents)
        self.actual_name = actual_name

    def __str__(self):
        return '<File %s>' % self.filename

    def dump(self):
        'Dump data to dict'
        return {
            '_id': self._id,
            'filename': self.filename,
            'checksum': self.checksum,
            'size': self.size,
            'actual_name': self.actual_name,
        }

    def delete(self, db):
        'Remove file'
        db.delete(self)


# Ref: https://stackoverflow.com/a/7524753
# Ref: https://stackoverflow.com/a/23734727
class InstructorAssociation(Base):
    'Relationship between instructor and course, many to many, with extra data'
    __tablename__ = 'instructor_assoc_table'
    left_id = Column(TEXT, ForeignKey('users.id'), primary_key=True)
    right_id = Column(TEXT, ForeignKey('courses.id'), primary_key=True)
    first_name = Column(TEXT)
    last_name = Column(TEXT)
    email = Column(TEXT)
    user = relationship(
        User,
        backref=backref(
            'inst_assoc', cascade='save-update, merge, delete, delete-orphan'
        ),
    )
    course = relationship(
        Course,
        backref=backref(
            'inst_assoc', cascade='save-update, merge, delete, delete-orphan'
        ),
    )

    @staticmethod
    def find(db, instructor, course):
        'Find association object from user and course'
        return (
            db.query(InstructorAssociation)
            .filter_by(user=instructor, course=course)
            .one_or_none()
        )

    def dump(self):
        'Dump data to dict'
        return {
            'left_id': self.left_id,
            'right_id': self.right_id,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'email': self.email,
        }


# Student -> Course (Many to Many)
class StudentAssociation(Base):
    'Relationship between student and course, many to many, with extra data'
    __tablename__ = 'student_assoc_table'
    left_id = Column(TEXT, ForeignKey('users.id'), primary_key=True)
    right_id = Column(TEXT, ForeignKey('courses.id'), primary_key=True)
    first_name = Column(TEXT)
    last_name = Column(TEXT)
    email = Column(TEXT)
    user = relationship(
        User,
        backref=backref(
            'student_assoc', cascade='save-update, merge, delete, delete-orphan'
        ),
    )
    course = relationship(
        Course,
        backref=backref(
            'student_assoc', cascade='save-update, merge, delete, delete-orphan'
        ),
    )

    @staticmethod
    def find(db, student, course):
        'Find association object from user and course'
        return (
            db.query(StudentAssociation)
            .filter_by(user=student, course=course)
            .one_or_none()
        )

    def dump(self):
        'Dump data to dict'
        return {
            'left_id': self.left_id,
            'right_id': self.right_id,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'email': self.email,
        }



'''
    ngshare Tornado server
'''

import os
import uuid
import json
import argparse
import base64
import binascii
import datetime
from urllib.parse import urlparse

from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import (
    Application,
    authenticated,
    RequestHandler,
    Finish,
    MissingArgumentError,
)
from jupyterhub.services.auth import HubAuthenticated
from sqlalchemy import create_engine, or_
from sqlalchemy.orm import sessionmaker

try:
    from . import dbutil
except ImportError:  # pragma: no cover
    import dbutil

try:
    from .database import (
        Base,
        User,
        Course,
        Assignment,
        Submission,
        File,
        InstructorAssociation,
        StudentAssociation,
        clear_db,
        init_db,
        dump_db,
    )
except ImportError:  # pragma: no cover
    from database import (
        Base,
        User,
        Course,
        Assignment,
        Submission,
        File,
        InstructorAssociation,
        StudentAssociation,
        clear_db,
        init_db,
        dump_db,
    )


class MyHelpers:
    'Helper functions for database accesses'

    def json_error(self, code, msg, **kwargs):
        'Abstract method resolved in MyRequestHandler'
        raise NotImplementedError

    def strftime(self, dt):
        'Use API specified format to strftime'
        return dt.strftime('%Y-%m-%d %H:%M:%S.%f %Z').strip()

    def strptime(self, string):
        'Use API specified format to strptime'
        try:
            return datetime.datetime.strptime(string, '%Y-%m-%d %H:%M:%S.%f %Z')
        except ValueError:
            try:
                return datetime.datetime.strptime(
                    string.strip(), '%Y-%m-%d %H:%M:%S.%f'
                )
            except ValueError:
                self.json_error(400, 'Time format incorrect')

    def path_check(self, pathname):
        '''
            Return whether a pathname (for file, in director tree) is safe
            Current policy:
                Not empty
                os.path.abspath resolves to a child address
                Path does not contain ('.', '..', '', '/')
            Note: os.path.abspath is used instead of os.path.realpath to prevent
             symbolic link issues, because the file is not on server
            Note: os.path.abspath is not 100% safe
            Note: currently only using Linux pathname conventions
        '''
        if not pathname:
            return False
        path = pathname
        while path:
            path, name = os.path.split(path)
            if name in ('.', '..', '', '/'):
                return False
        working = os.path.abspath('.')
        target = os.path.abspath(pathname)
        if os.path.commonpath([working, target]) != working:
            return False
        return True

    def filename_create(self, filename):
        'Create a file name for storage; trying to follow extension'
        ext = os.path.splitext(filename)[1]
        if len(ext) > 10:
            ext = ''
        return str(uuid.uuid4()) + ext

    def json_files_pack(self, file_list, list_only):
        'Generate JSON directory tree from a list of File objects'
        ans = []
        for i in file_list:
            entry = {
                'path': i.filename,
                'checksum': i.checksum,
            }
            if not list_only:
                storage_path = self.application.storage_path
                actual_path = os.path.join(storage_path, i.actual_name)
                content = open(actual_path, 'rb').read()
                entry['content'] = base64.encodebytes(content).decode()
            ans.append(entry)
        return ans

    def json_files_unpack(self, json_str, target):
        '''
            Generate a list of File objects from a JSON directory tree
            json_str: json object as string; raise error when None
            target: a list to put file objects in
        '''
        if json_str is None:
            self.json_error(400, 'Please supply files')
        try:
            json_obj = json.loads(json_str)
        except json.decoder.JSONDecodeError:
            self.json_error(400, 'Files cannot be JSON decoded')
        content_list = []
        for i in json_obj:
            if not self.path_check(i['path']):
                self.json_error(400, 'Illegal path')
            try:
                content = base64.decodebytes(i['content'].encode())
            except binascii.Error:
                self.json_error(400, 'Content cannot be base64 decoded')
            target.append(File(i['path'], content))
            content_list.append(content)
        # Commit files
        storage_path = self.application.storage_path
        os.makedirs(storage_path, exist_ok=True)
        for file_obj, content in zip(target, content_list):
            f = None
            for i in range(10):
                actual_name = self.filename_create(file_obj.filename)
                try:
                    f = open(os.path.join(storage_path, actual_name), 'xb')
                    break
                except FileExistsError:
                    pass
            if f is None:
                raise self.json_error(500, 'Internal server error')
            f.write(content)
            f.close()
            file_obj.actual_name = actual_name

    def find_or_create_user(self, user_id):
        'Return a User object from id; create if not found'
        user = self.db.query(User).filter(User.id == user_id).one_or_none()
        if user is None:
            user = User(user_id)
            self.db.add(user)
            self.db.commit()
        return user

    def find_course(self, course_id):
        'Return a Course object from id, or raise error'
        qry = self.db.query(Course)
        course = qry.filter(Course.id == course_id).one_or_none()
        if course is None:
            self.json_error(404, 'Course not found')
        return course

    def find_assignment(self, course, assignment_id):
        'Return an Assignment object from course and id, or raise error'
        assignment = (
            self.db.query(Assignment)
            .filter(Assignment.id == assignment_id, Assignment.course == course)
            .one_or_none()
        )
        if assignment is None:
            self.json_error(404, 'Assignment not found')
        return assignment

    def find_course_instructor(self, course, instructor_id):
        'Return a instructor as User object from course and id'
        instructor = (
            self.db.query(User)
            .filter(User.id == instructor_id, User.teaching.contains(course))
            .one_or_none()
        )
        if instructor is None:
            self.json_error(404, 'Instructor not found')
        return instructor

    def find_course_student(self, course, student_id):
        'Return a student as User object from course and id'
        student = (
            self.db.query(User)
            .filter(User.id == student_id, User.taking.contains(course))
            .one_or_none()
        )
        if student is None:
            self.json_error(404, 'Student not found')
        return student

    def find_course_user(self, course, user_id):
        'Return a student or instructor as User object from course and id'
        user = (
            self.db.query(User)
            .filter(
                User.id == user_id,
                or_(
                    User.taking.contains(course), User.teaching.contains(course)
                ),
            )
            .one_or_none()
        )
        if user is None:
            self.json_error(404, 'Student not found')
        return user

    def find_student_submissions(self, assignment, student):
        'Return a list of Submission objects from assignment and student'
        return self.db.query(Submission).filter(
            Submission.assignment == assignment, Submission.student == student
        )

    def find_student_latest_submission(self, assignment, student):
        'Return the latest Submission object from assignment and studnet'
        submission = (
            self.find_student_submissions(assignment, student)
            .order_by(Submission.timestamp.desc())
            .first()
        )
        if submission is None:
            self.json_error(404, 'Submission not found')
        return submission

    def find_student_submission(self, assignment, student, timestamp):
        'Return the Submission object from timestamp etc'
        submission = (
            self.find_student_submissions(assignment, student)
            .filter(Submission.timestamp == timestamp)
            .one_or_none()
        )
        if submission is None:
            self.json_error(404, 'Submission not found')
        return submission

    # User management

    def wrap_instructor_info(self, user, course):
        'Return dict of instructor info (full name, email, etc)'
        association = InstructorAssociation.find(self.db, user, course)
        if association is None:  # Error in data integrity
            return {
                'username': user.id,
                'first_name': None,
                'last_name': None,
                'email': None,
            }
        return {
            'username': user.id,
            'first_name': association.first_name,
            'last_name': association.last_name,
            'email': association.email,
        }

    def wrap_student_info(self, user, course):
        'Return dict of student info (full name, email, etc)'
        association = StudentAssociation.find(self.db, user, course)
        if association is None:  # Error in data integrity
            return {
                'username': user.id,
                'first_name': None,
                'last_name': None,
                'email': None,
            }
        return {
            'username': user.id,
            'first_name': association.first_name,
            'last_name': association.last_name,
            'email': association.email,
        }

    # Auth APIs

    def is_admin(self):
        'Return whether user is admin user'
        return self.user.id in self.application.admin

    def is_course_student(self, course, user):
        'Return whether user is a student in the course'
        return course in user.taking

    def is_course_instructor(self, course, user):
        'Return whether user is an instructor in the course'
        return course in user.teaching

    def check_admin(self):
        'Assert user is admin user'
        if not self.is_admin():
            msg = 'Permission denied'
            if self.application.debug:
                msg += ' (not admin)'
            self.json_error(403, msg)

    def check_course_instructor(self, course):
        'Assert user is an instructor in the course, or admin'
        if not self.is_admin() and not self.is_course_instructor(
            course, self.user
        ):
            msg = 'Permission denied'
            if self.application.debug:
                msg += ' (not course instructor)'
            self.json_error(403, msg)

    def check_course_user(self, course):
        'Assert user is a student or an instructor in the course, or admin'
        if (
            not self.is_admin()
            and not self.is_course_instructor(course, self.user)
            and not self.is_course_student(course, self.user)
        ):
            msg = 'Permission denied'
            if self.application.debug:
                msg += ' (not related to course)'
            self.json_error(403, msg)


class MyRequestHandler(HubAuthenticated, RequestHandler, MyHelpers):
    'Custom request handler for ngshare'

    def json_success(self, msg=None, **kwargs):
        'Return success as a JSON object'
        assert 'success' not in kwargs and 'message' not in kwargs
        resp = {'success': True, **kwargs}
        if msg is not None:
            resp['message'] = msg
        raise Finish(json.dumps(resp))

    def json_error(self, code, msg, **kwargs):
        'Return error as a JSON object'
        assert 'success' not in kwargs and 'message' not in kwargs
        self.set_status(code)
        raise Finish(json.dumps({'success': False, 'message': msg, **kwargs}))

    def prepare(self):
        'Provide a db object'
        self.db = self.application.db_session()
        current_user = self.get_current_user()
        if current_user is not None:
            self.user = User.from_jupyterhub_user(current_user, self.db)
        else:
            self.user = None

    def on_finish(self):
        self.db.close()


class HomePage(MyRequestHandler):
    '/api/'

    @authenticated
    def get(self):
        'Display an HTML page for debugging'
        self.render(
            'home.html',
            vngshare=self.application.vngshare,
            debug=self.application.debug or self.is_admin(),
        )


class Static(MyRequestHandler):
    '/api/favicon.ico, /api/masonry.min.js'

    @authenticated
    def get(self, name):
        'Static files'
        pwd = os.path.dirname(os.path.realpath(__file__))
        file_name = os.path.join(pwd, name)
        self.write(open(file_name, 'rb').read())


class ListCourses(MyRequestHandler):
    '/api/courses'

    @authenticated
    def get(self):
        '''
            List all available courses the user is taking or teaching. (anyone)
            List all courses in ngshare. (admin)
        '''
        courses = set()
        if self.is_admin():
            for i in self.db.query(Course).all():
                courses.add(i.id)
        else:
            for i in self.user.teaching:
                courses.add(i.id)
            for i in self.user.taking:
                courses.add(i.id)
        self.json_success(courses=sorted(courses))


class AddCourse(MyRequestHandler):
    '/api/course/<course_id>'

    @authenticated
    def post(self, course_id):
        'Add a course (admin)'
        self.check_admin()
        if self.db.query(Course).filter(Course.id == course_id).one_or_none():
            self.json_error(409, 'Course already exists')
        try:
            instructors = []
            for i in json.loads(self.get_argument('instructors', '[]')):
                instructors.append(self.find_or_create_user(i))
        except json.decoder.JSONDecodeError:
            self.json_error(400, 'Instructors cannot be JSON decoded')
        course = Course(course_id, instructors)
        self.db.add(course)
        self.db.commit()
        self.json_success()

    @authenticated
    def delete(self, course_id):
        'Remove a course (admin)'
        self.check_admin()
        course = self.find_course(course_id)
        course.delete(self.db)
        self.db.commit()
        self.json_success()


class ManageInstructor(MyRequestHandler):
    '/api/instructor/<course_id>/<instructor_id>'

    @authenticated
    def post(self, course_id, instructor_id):
        '''
            Add or update a course instructor. (admin)
            Update self full name or email. (instructors)
        '''
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        instructor = self.find_or_create_user(instructor_id)
        first_name = self.get_argument('first_name', None)
        if first_name is None:
            self.json_error(400, 'Please supply first name')
        last_name = self.get_argument('last_name', None)
        if last_name is None:
            self.json_error(400, 'Please supply last name')
        email = self.get_argument('email', None)
        if email is None:
            self.json_error(400, 'Please supply email')
        if instructor in course.students:
            if not self.is_admin():
                self.json_error(
                    400, 'Permission denied' ' (cannot modify instructors)'
                )
            course.students.remove(instructor)
        if instructor not in course.instructors:
            if not self.is_admin():
                self.json_error(
                    400, 'Permission denied' ' (cannot modify instructors)'
                )
            course.instructors.append(instructor)
        if not self.is_admin() and instructor.id != self.user.id:
            self.json_error(
                400, 'Permission denied' ' (cannot modify other instructors)'
            )
        association = InstructorAssociation.find(self.db, instructor, course)
        association.first_name = first_name
        association.last_name = last_name
        association.email = email
        self.db.commit()
        self.json_success()

    @authenticated
    def get(self, course_id, instructor_id):
        'Get information about a course instructor. (instructors+students)'
        course = self.find_course(course_id)
        self.check_course_user(course)
        instructor = self.find_course_instructor(course, instructor_id)
        ans = self.wrap_instructor_info(instructor, course)
        self.json_success(**ans)

    @authenticated
    def delete(self, course_id, instructor_id):
        'Remove a course instructor (admin)'
        self.check_admin()
        course = self.find_course(course_id)
        instructor = self.find_course_instructor(course, instructor_id)
        course.instructors.remove(instructor)
        self.db.commit()
        self.json_success()


class ListInstructors(MyRequestHandler):
    '/api/instructors/<course_id>/'

    @authenticated
    def get(self, course_id):
        'Get information about all course instructors. (instructors+students)'
        course = self.find_course(course_id)
        self.check_course_user(course)
        ans = []
        for instructor in course.instructors:
            ans.append(self.wrap_instructor_info(instructor, course))
        self.json_success(instructors=ans)


class ManageStudent(MyRequestHandler):
    '/api/student/<course_id>/<student_id>'

    @authenticated
    def post(self, course_id, student_id):
        'Add or update a student. (instructors only)'
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        student = self.find_or_create_user(student_id)
        first_name = self.get_argument('first_name', None)
        if first_name is None:
            self.json_error(400, 'Please supply first name')
        last_name = self.get_argument('last_name', None)
        if last_name is None:
            self.json_error(400, 'Please supply last name')
        email = self.get_argument('email', None)
        if email is None:
            self.json_error(400, 'Please supply email')
        if student in course.instructors:
            self.json_error(409, 'Cannot add instructor as student')
        if student not in course.students:
            course.students.append(student)
        association = StudentAssociation.find(self.db, student, course)
        association.first_name = first_name
        association.last_name = last_name
        association.email = email
        self.db.commit()
        self.json_success()

    @authenticated
    def get(self, course_id, student_id):
        '''
            Get information about a student.
            (instructors+student with same student_id)
        '''
        course = self.find_course(course_id)
        if self.user.id != student_id:
            self.check_course_instructor(course)
        student = self.find_course_student(course, student_id)
        ans = self.wrap_student_info(student, course)
        self.json_success(**ans)

    @authenticated
    def delete(self, course_id, student_id):
        'Remove a student (instructors only)'
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        student = self.find_course_student(course, student_id)
        course.students.remove(student)
        self.db.commit()
        self.json_success()


class ListStudents(MyRequestHandler):
    '/api/students/<course_id>/'

    @authenticated
    def post(self, course_id):
        'Add or update students. (instructors only)'
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        students_str = self.get_argument('students', None)
        # Check request format
        if not students_str:
            self.json_error(400, 'Please supply students')
        try:
            students = json.loads(students_str)
        except json.decoder.JSONDecodeError:
            self.json_error(400, 'Students cannot be JSON decoded')
        if not isinstance(students, list):
            self.json_error(400, 'Incorrect request format')
        if not students:
            self.json_error(400, 'Please supply students')
        for i in students:
            if (
                not isinstance(i, dict)
                or not isinstance(i.get('username'), str)
                or not isinstance(i.get('first_name'), str)
                or not isinstance(i.get('last_name'), str)
                or not isinstance(i.get('email'), str)
            ):
                self.json_error(400, 'Incorrect request format')
        # Commit
        ans = []
        for i in students:
            student = self.find_or_create_user(i['username'])
            if student in course.instructors:
                ans.append(
                    {
                        'username': i['username'],
                        'success': False,
                        'message': 'Cannot add instructor as student',
                    }
                )
                continue
            if student not in course.students:
                course.students.append(student)
            association = StudentAssociation.find(self.db, student, course)
            association.first_name = i['first_name']
            association.last_name = i['last_name']
            association.email = i['email']
            ans.append(
                {'username': i['username'], 'success': True,}
            )
        self.db.commit()
        self.json_success(status=ans)

    @authenticated
    def get(self, course_id):
        'Get information about all course students. (instructors only)'
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        ans = []
        for student in course.students:
            ans.append(self.wrap_student_info(student, course))
        self.json_success(students=ans)


class ListAssignments(MyRequestHandler):
    '/api/assignments/<course_id>'

    @authenticated
    def get(self, course_id):
        'List all assignments for a course (students+instructors)'
        course = self.find_course(course_id)
        self.check_course_user(course)
        assignments = course.assignments
        self.json_success(assignments=list(map(lambda x: x.id, assignments)))


class DownloadReleaseAssignment(MyRequestHandler):
    '/api/assignment/<course_id>/<assignment_id>'

    def get(self, course_id, assignment_id):
        'Download a copy of an assignment (students+instructors)'
        course = self.find_course(course_id)
        self.check_course_user(course)
        assignment = self.find_assignment(course, assignment_id)
        list_only = self.get_argument('list_only', 'false') == 'true'
        files = self.json_files_pack(assignment.files, list_only)
        self.json_success(files=files)

    def post(self, course_id, assignment_id):
        'Release an assignment (instructors only)'
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        if (
            self.db.query(Assignment)
            .filter(Assignment.id == assignment_id, Assignment.course == course)
            .one_or_none()
        ):
            self.json_error(409, 'Assignment already exists')
        assignment = Assignment(assignment_id, course)
        files = self.get_argument('files', None)
        self.json_files_unpack(files, assignment.files)
        self.db.commit()
        self.json_success()

    def delete(self, course_id, assignment_id):
        'Remove an assignment (instructors only)'
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        assignment = self.find_assignment(course, assignment_id)
        assignment.delete(self.db)
        self.db.commit()
        self.json_success()


class ListSubmissions(MyRequestHandler):
    '/api/submissions/<course_id>/<assignment_id>'

    def get(self, course_id, assignment_id):
        '''
            List all submissions for an assignment from all students
             (instructors only)
        '''
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        assignment = self.find_assignment(course, assignment_id)
        submissions = []
        for submission in assignment.submissions:
            submissions.append(
                {
                    'student_id': submission.student.id,
                    'timestamp': self.strftime(submission.timestamp),
                }
            )
        self.json_success(submissions=submissions)


class ListStudentSubmissions(MyRequestHandler):
    '/api/submissions/<course_id>/<assignment_id>/<student_id>'

    def get(self, course_id, assignment_id, student_id):
        '''
            List all submissions for an assignment from a particular student
             (instructors+students,
              students restricted to their own submissions)
        '''
        course = self.find_course(course_id)
        if self.user.id != student_id:
            self.check_course_instructor(course)
        assignment = self.find_assignment(course, assignment_id)
        student = self.find_course_user(course, student_id)
        submissions = []
        for submission in self.find_student_submissions(assignment, student):
            submissions.append(
                {
                    'student_id': submission.student.id,
                    'timestamp': self.strftime(submission.timestamp),
                }
            )
        self.json_success(submissions=submissions)


class SubmitAssignment(MyRequestHandler):
    '/api/submission/<course_id>/<assignment_id>'

    def post(self, course_id, assignment_id):
        'Submit a copy of an assignment (students+instructors)'
        course = self.find_course(course_id)
        self.check_course_user(course)
        assignment = self.find_assignment(course, assignment_id)
        submission = Submission(self.user, assignment)
        files = self.get_body_argument('files', None)
        self.json_files_unpack(files, submission.files)
        self.db.commit()
        self.json_success(timestamp=self.strftime(submission.timestamp))


class DownloadAssignment(MyRequestHandler):
    '/api/submission/<course_id>/<assignment_id>/<student_id>'

    def get(self, course_id, assignment_id, student_id):
        '''
            Download a student's submitted assignment (instructors only)
            TODO: maybe allow student to see their own submissions?
        '''
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        assignment = self.find_assignment(course, assignment_id)
        student = self.find_course_user(course, student_id)
        list_only = self.get_argument('list_only', 'false') == 'true'
        timestamp = self.get_argument('timestamp', '')
        if not timestamp:
            submission = self.find_student_latest_submission(
                assignment, student
            )
        else:
            submission = self.find_student_submission(
                assignment, student, timestamp
            )
        files = self.json_files_pack(submission.files, list_only)
        self.json_success(
            files=files, timestamp=self.strftime(submission.timestamp)
        )


class UploadDownloadFeedback(MyRequestHandler):
    '/api/feedback/<course_id>/<assignment_id>/<student_id>'

    def post(self, course_id, assignment_id, student_id):
        '''
            POST /api/feedback/<course_id>/<assignment_id>/<student_id>
            Upload feedback on a student's assignment (instructors only)
        '''
        course = self.find_course(course_id)
        self.check_course_instructor(course)
        assignment = self.find_assignment(course, assignment_id)
        student = self.find_course_user(course, student_id)
        try:
            timestamp = self.strptime(self.get_body_argument('timestamp'))
        except MissingArgumentError:
            self.json_error(400, 'Please supply timestamp')
        submission = self.find_student_submission(
            assignment, student, timestamp
        )
        for file_obj in submission.feedbacks:
            file_obj.delete(self.db)
        submission.feedbacks.clear()
        files = self.get_body_argument('files', None)
        self.json_files_unpack(files, submission.feedbacks)
        self.db.commit()
        self.json_success()

    def get(self, course_id, assignment_id, student_id):
        '''
            GET /api/feedback/<course_id>/<assignment_id>/<student_id>
            Download feedback on a student's assignment
             (instructors+students, students restricted to own submissions)
        '''
        course = self.find_course(course_id)
        if self.user.id != student_id:
            self.check_course_instructor(course)
        assignment = self.find_assignment(course, assignment_id)
        student = self.find_course_user(course, student_id)
        try:
            timestamp = self.strptime(self.get_query_argument('timestamp'))
        except MissingArgumentError:
            self.json_error(400, 'Please supply timestamp')
        submission = self.find_student_submission(
            assignment, student, timestamp
        )
        list_only = self.get_argument('list_only', 'false') == 'true'
        files = self.json_files_pack(submission.feedbacks, list_only)
        self.json_success(
            files=files, timestamp=self.strftime(submission.timestamp)
        )


class InitDatabase(MyRequestHandler):
    '/initialize-Data6ase'

    @authenticated
    def get(self):
        'Initialize database similar to in vserver'
        # Dangerous: do not use in production
        action = self.get_argument('action', None)
        if not self.application.debug:
            if not self.is_admin() or action != 'dump':
                self.json_error(403, 'Debug mode is off')
        if action == 'clear':
            clear_db(self.db, self.application.storage_path)
            self.json_success('done')
        elif action == 'init':
            init_db(self.db, self.application.storage_path)
            self.json_success('done')
        elif action == 'dump':
            result = dump_db(self.db)
            if self.get_argument('human-readable', 'false') != 'true':
                self.json_success(**result)
            ans = []
            for key, value in result.items():
                thead = list(value[0])
                tbody = []
                for line in value:
                    tbody.append(list(map(line.__getitem__, thead)))
                ans.append(
                    {'header': key, 'thead': thead, 'tbody': tbody,}
                )
            self.render('dump.html', tables=ans)
        else:
            self.json_error(400, 'action should be clear, init, or dump')


class NotFoundHandler(RequestHandler):
    '404 handler'

    def get(self):
        'Disable 404 page'
        self.set_status(404)
        self.write("<h1>404 Not Found</h1>\n")
        if not self.application.debug:
            return
        self.write(json.dumps(dict(os.environ), indent=1, sort_keys=True))
        self.write('\n' + self.request.uri + '\n' + self.request.path + '\n')


class HealthCheckHandler(RequestHandler):
    '/healthz'

    def get(self):
        "Health check endpoint for k8s. Returns a 200 to indicate we're alive"
        self.write(json.dumps({'success': True}))
        return


class MyApplication(Application):
    'Custom application for ngshare'

    def __init__(
        self,
        prefix,
        db_url,
        storage_path,
        debug=False,
        admin=(),
        autoreload=True,
    ):

        if prefix.startswith('/healthz/'):
            raise ValueError("API prefix may not start with /healthz/")

        handlers = [
            (prefix, HomePage),
            (prefix + r'(favicon\.ico)', Static),
            (prefix + r'(masonry\.min\.js)', Static),
            (prefix + 'courses', ListCourses),
            (prefix + 'course/([^/]+)', AddCourse),
            (prefix + 'instructor/([^/]+)/([^/]+)', ManageInstructor),
            (prefix + 'instructors/([^/]+)', ListInstructors),
            (prefix + 'student/([^/]+)/([^/]+)', ManageStudent),
            (prefix + 'students/([^/]+)', ListStudents),
            (prefix + 'assignments/([^/]+)', ListAssignments),
            (prefix + 'assignment/([^/]+)/([^/]+)', DownloadReleaseAssignment),
            (prefix + 'submissions/([^/]+)/([^/]+)', ListSubmissions),
            (
                prefix + 'submissions/([^/]+)/([^/]+)/([^/]+)',
                ListStudentSubmissions,
            ),
            (prefix + 'submission/([^/]+)/([^/]+)', SubmitAssignment),
            (prefix + 'submission/([^/]+)/([^/]+)/([^/]+)', DownloadAssignment),
            (
                prefix + 'feedback/([^/]+)/([^/]+)/([^/]+)',
                UploadDownloadFeedback,
            ),
            (prefix + 'initialize-Data6ase', InitDatabase),
            ('/healthz', HealthCheckHandler),
        ]
        handlers.append((r'.*', NotFoundHandler))
        super(MyApplication, self).__init__(
            handlers, debug=debug, autoreload=autoreload
        )
        # Connect Database
        engine = create_engine(db_url)
        Base.metadata.bind = engine
        Base.metadata.create_all(engine)
        self.db_session = sessionmaker(bind=engine)
        self.storage_path = storage_path
        self.debug = debug
        self.vngshare = False
        self.admin = admin


class MockAuth(HubAuthenticated):
    '''
        Mock class substituting HubAuthenticated, for vngshare
    '''

    def get_login_url(self):
        return 'http://example.com/'

    def get_current_user(self):
        if type(self).__name__ in ('HomePage', 'Static', 'InitDatabase'):
            user = self.get_argument('user', 'user')
        else:
            user = self.get_argument('user')
        return {'name': user}


def main():  # pragma: no cover
    'Main function'
    parser = argparse.ArgumentParser(
        description='ngshare, a REST API nbgrader exchange'
    )
    parser.add_argument(
        '--vngshare',
        help='Use vngshare (stand-alone mode)',
        action='store_true',
    )
    parser.add_argument(
        '--jupyterhub_api_url', help='override $JUPYTERHUB_API_URL'
    )
    parser.add_argument(
        '--prefix',
        help='URL prefix (default: $JUPYTERHUB_SERVICE_PREFIX or "/api/")',
    )
    parser.add_argument('--debug', action='store_true', help='enable debug')
    parser.add_argument(
        '--database',
        help='database url',
        default='sqlite:////srv/ngshare/ngshare.db',
    )
    parser.add_argument(
        '--storage', help='path to store files', default='/srv/ngshare/files/'
    )
    parser.add_argument(
        '--admins', help='admin user ids (comma splitted)', default=''
    )
    parser.add_argument(
        '--host', help='bind hostname (vngshare only)', default='127.0.0.1'
    )
    parser.add_argument(
        '--port', help='bind port (vngshare only)', type=int, default=12121
    )
    parser.add_argument(
        '--no-upgrade-db',
        action='store_true',
        help='do not automatically upgrade database',
    )
    args = parser.parse_args()
    if args.jupyterhub_api_url is not None:
        os.environ['JUPYTERHUB_API_URL'] = args.jupyterhub_api_url

    if not args.no_upgrade_db:
        dbutil.upgrade(args.database)

    prefix = args.prefix or os.environ.get('JUPYTERHUB_SERVICE_PREFIX', '/api/')

    if args.vngshare:
        MyRequestHandler.__bases__ = (MockAuth, RequestHandler, MyHelpers)

    app = MyApplication(
        prefix,
        args.database,
        args.storage,
        admin=args.admins.split(','),
        debug=args.debug,
    )

    http_server = HTTPServer(app)

    if args.vngshare:
        http_server.listen(args.port, args.host)
        app.vngshare = True
        print()
        print('Starting vngshare (Vserver-like Notebook Grader Share)')
        print('DO NOT USE IN PRODUCTION')
        print('Database file is %s' % repr(args.database))
        print('Storage directory is %s' % repr(args.storage))
        print('admin users are %s' % repr(app.admin))
        print('Please go to http://%s:%d/api/' % (args.host, args.port))
    else:
        url = urlparse(os.environ['JUPYTERHUB_SERVICE_URL'])

        # Must listen on all interfaces for proxy
        http_server.listen(url.port, '0.0.0.0')

    IOLoop.current().start()


if __name__ == '__main__':  # pragma: no cover
    main()



'''
    Test database migration tools
'''

import os
import tempfile

from . import dbutil


def test_update():
    'Test update database'
    # Create file name
    tempdb_path = tempfile.mktemp('.db')
    tempdb_url = 'sqlite:///' + tempdb_path
    # Create database
    assert not os.path.exists(tempdb_path)
    dbutil.upgrade(tempdb_url)
    # Downgrade to init
    dbutil.main(['downgrade', 'aa00db20c10a'], tempdb_url)
    # Downgrade to nothing
    dbutil.main(['downgrade', 'base'], tempdb_url)
    # Offline mode
    dbutil.main(['upgrade', 'head', '--sql'], tempdb_url)
    # Upgrade to head
    dbutil.main(['upgrade', 'head'], tempdb_url)
    # Invalid argument error
    try:
        dbutil.main([])
    except SystemExit:
        pass
    # Remove tempdb
    os.remove(tempdb_path)



"""Database utilities for nbgrader"""
# Based on jupyterhub.dbutil
# Based on nbgrader.dbutil

import os
import sys

import alembic
import alembic.command
import alembic.config

_here = os.path.abspath(os.path.dirname(__file__))

ALEMBIC_INI_TEMPLATE_PATH = os.path.join(_here, 'alembic.ini')
ALEMBIC_DIR = os.path.join(_here, 'alembic')

DEFAULT_DB = 'sqlite:////tmp/ngshare.db'


def get_alembic_config(db_url: str = DEFAULT_DB) -> alembic.config.Config:
    """Generate the alembic configuration from the template and populate fields.
    db_url: str [default: 'sqlite:////tmp/ngshare.db']
        The database url used to populate sqlalchemy.url, e.g. `sqlite:///ngshare.db`.
    """

    config = alembic.config.Config(ALEMBIC_INI_TEMPLATE_PATH)
    config.set_main_option("script_location", ALEMBIC_DIR)
    config.set_main_option("sqlalchemy.url", db_url)
    return config


def upgrade(db_url: str = DEFAULT_DB, revision='head'):
    """Upgrade the given database to revision.
    db_url: str [default: 'sqlite:////tmp/ngshare.db']
        The SQLAlchemy database url, e.g. `sqlite:///ngshare.db`.
    revision: str [default: head]
        The alembic revision to upgrade to.
    """
    alembic.command.upgrade(get_alembic_config(db_url), revision)


def main(args: list, db_url: str = DEFAULT_DB):
    """Run an alembic command with the right config"""
    cl = alembic.config.CommandLine()
    options = cl.parser.parse_args(args)
    if not hasattr(options, "cmd"):
        cl.parser.error("too few arguments")
    cl.run_cmd(get_alembic_config(db_url), options)


if __name__ == '__main__':  # pragma: no cover
    main(sys.argv[1:])



from .database import (
    Base,
    User,
    Course,
    Assignment,
    Submission,
    File,
    InstructorAssociation,
    StudentAssociation,
)
from .test_database import clear_db, init_db, dump_db

__all__ = [
    'Base',
    'User',
    'Course',
    'Assignment',
    'Submission',
    'File',
    'InstructorAssociation',
    'StudentAssociation',
    'clear_db',
    'init_db',
    'dump_db',
]



'''
    vngshare - Vserver-like Notebook Grader Share
    Similar to vserver; allows easy testing.
'''

import os
import sys

if __name__ == '__main__':  # pragma: no cover
    os.execvp(
        'python3',
        [
            'python3',
            'ngshare.py',
            '--vngshare',
            '--database',
            'sqlite:////tmp/ngshare.db',
            '--storage',
            '/tmp/ngshare',
            '--debug',
        ]
        + sys.argv[1:],
    )



'''
    Test database structure and some properties of SQLAlchemy
'''

import os
import shutil
import tempfile

from collections import defaultdict

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

Session = None
test_storage = None

from .database import *


def clear_db(db, storage_path):
    'Remove all data from database'
    db.query(User).delete()
    db.query(Course).delete()
    db.query(Assignment).delete()
    db.query(Submission).delete()
    db.query(File).delete()
    db.query(InstructorAssociation).delete()
    db.query(StudentAssociation).delete()
    for table_name in [
        'assignment_files_assoc_table',
        'submission_files_assoc_table',
        'feedback_files_assoc_table',
    ]:
        db.execute('DELETE FROM %s' % table_name)
    db.commit()
    if storage_path is not None:
        shutil.rmtree(storage_path, ignore_errors=True)


def init_db(db, storage_path):
    '''
        Create testing data
        course1
            instructor = [kevin]
            student = [lawrence]
            assignments = [challenge] (two submissions, one feedback)
        course2
            instructor = [abigail]
            student = [eric]
            assignments = [assignment2a, assignment2b] (no submissions)
    '''
    uk = User('kevin')
    ua = User('abigail')
    ul = User('lawrence')
    ue = User('eric')
    course1 = Course('course1', [uk])
    course2 = Course('course2', [ua])
    db.add(course1)
    db.add(course2)
    course1.students.append(ul)
    course2.students.append(ue)
    aa = Assignment('assignment2a', course2)
    ab = Assignment('assignment2b', course2)
    ac = Assignment('challenge', course1)
    db.add(aa)
    db.add(ab)
    db.add(ac)
    s1 = Submission(ul, ac)
    s2 = Submission(ul, ac)
    s1.timestamp = datetime.datetime(2020, 1, 1, 0, 0, 0, 0)
    db.add(s1)
    db.add(s2)
    aa.files.append(File('file0', b'00000', 'actual0'))
    ab.files.append(File('file1', b'11111', 'actual1'))
    ac.files.append(File('file2', b'22222', 'actual2'))
    s1.files.append(File('file3', b'33333', 'actual3'))
    s2.files.append(File('file4', b'44444', 'actual4'))
    s1.feedbacks.append(File('file5', b'55555', 'actual5'))
    os.makedirs(storage_path, exist_ok=True)
    for i in range(6):
        f = open(os.path.join(storage_path, 'actual%d' % i), 'wb')
        f.write((b'%d' % i) * 5)
        f.close()
    db.commit()


def dump_db(db):
    'Dump database out'
    ans = defaultdict(list)
    for table in (
        User,
        Course,
        Assignment,
        Submission,
        File,
        InstructorAssociation,
        StudentAssociation,
    ):
        for i in db.query(table).all():
            ans[table.__tablename__].append(i.dump())
    for table in (
        assignment_files_assoc_table,
        submission_files_assoc_table,
        feedback_files_assoc_table,
    ):
        for i in db.query(table).all():
            ans[table.name].append(
                {'left_id': i.left_id, 'right_id': i.right_id,}
            )
    return ans


def test_legacy():
    'Some test cases created when building database structure'
    global Session
    engine = create_engine('sqlite://')  # temp database in memory
    Base.metadata.bind = engine
    Base.metadata.create_all(engine)
    Session = sessionmaker(bind=engine)
    db = Session()

    # populate with random data
    instructor = User("rkevin")
    student = User("rk-test")
    db.add(instructor)
    db.add(student)
    course = Course("ecs189m", [instructor])
    assignment = Assignment("ps1", course)
    submission = Submission(student, assignment)
    db.add(course)
    db.commit()
    print(submission)

    # check if values are sane
    print("List of users:")
    for i in db.query(User):
        print(i.id)
    print()

    print("User rkevin is teaching the courses:")
    user = db.query(User).filter(User.id == "rkevin").one_or_none()
    for i in user.teaching:
        print(i.id)

    print("In the 'ecs189m' class, there are these assignments:")
    course = db.query(Course).filter(Course.id == "ecs189m").one_or_none()
    for i in course.assignments:
        print("Assignment name:", i.id)
        print("There are", len(i.submissions), "submissions")
        for j in i.submissions:
            print("One submission from", j.student, "at", j.timestamp)


def test_init():
    'Test clearing database and fill in default test data'
    global Session, test_storage
    test_storage = tempfile.mkdtemp()
    db = Session()
    clear_db(db, None)
    assert not db.query(assignment_files_assoc_table).all()
    assert not db.query(submission_files_assoc_table).all()
    assert not db.query(feedback_files_assoc_table).all()
    assert not db.query(User).all()
    assert not db.query(Course).all()
    assert not db.query(Assignment).all()
    assert not db.query(Submission).all()
    assert not db.query(File).all()
    assert not db.query(InstructorAssociation).all()
    assert not db.query(StudentAssociation).all()
    assert not dump_db(db)
    init_db(db, test_storage)
    assert len(db.query(User).all()) == 4
    assert len(db.query(Course).all()) == 2
    assert len(db.query(Assignment).all()) == 3
    assert len(db.query(Submission).all()) == 2
    assert len(db.query(File).all()) == 6
    dumped = dump_db(db)
    assert len(dumped['users']) == 4
    assert len(dumped['courses']) == 2
    assert len(dumped['assignments']) == 3
    assert len(dumped['submissions']) == 2
    assert len(dumped['files']) == 6


def test_upload_feedback():
    'When uploading feedback, old feedbacks need to be removed'
    global Session
    db = Session()
    ts = datetime.datetime(2020, 1, 1, 0, 0, 0, 0)
    s1 = db.query(Submission).filter(Submission.timestamp == ts).one_or_none()
    assert s1 is not None
    assert len(s1.feedbacks) == 1
    for file in s1.feedbacks:
        db.delete(file)
    s1.feedbacks.clear()
    db.commit()
    assert len(db.query(File).all()) == 5


def test_remove_assignment():
    'Test when removing assignment, submissions and files need to be removed'
    global Session
    db = Session()
    ac = db.query(Assignment).filter(Assignment.id == 'challenge').one_or_none()
    assert ac is not None
    ac.delete(db)
    db.commit()
    assert len(db.query(Assignment).all()) == 2
    assert len(db.query(Submission).all()) == 0
    assert len(db.query(File).all()) == 2


def test_instructor_association():
    'Test instructor association table'
    global Session
    db = Session()
    kevin = db.query(User).filter_by(id='kevin').one_or_none()
    assert kevin.id == 'kevin'
    assert len(kevin.teaching) == 1
    course1 = kevin.teaching[0]
    assert course1.id == 'course1'
    relation_count = len(db.query(InstructorAssociation).all())
    # Check relation
    kevin.teaching.remove(kevin.teaching[0])
    db.commit()
    assert len(course1.instructors) == 0
    assert len(db.query(InstructorAssociation).all()) == relation_count - 1
    course1.instructors.append(kevin)
    db.commit()
    assert len(db.query(InstructorAssociation).all()) == relation_count
    assert len(kevin.teaching) == 1
    # Check data
    association = InstructorAssociation.find(db, kevin, course1)
    assert association.first_name is None
    assert association.last_name is None
    assert association.email is None
    association.first_name = 'Kevin.first_name'
    association.last_name = 'Kevin.last_name'
    association.email = 'Kevin.email'
    db.commit()
    association = InstructorAssociation.find(db, kevin, course1)
    assert association.first_name == 'Kevin.first_name'
    assert association.last_name == 'Kevin.last_name'
    assert association.email == 'Kevin.email'


def test_student_association():
    'Test student association table'
    global Session
    db = Session()
    lawrence = db.query(User).filter_by(id='lawrence').one_or_none()
    assert lawrence.id == 'lawrence'
    assert len(lawrence.taking) == 1
    course1 = lawrence.taking[0]
    assert course1.id == 'course1'
    relation_count = len(db.query(StudentAssociation).all())
    # Check relation
    lawrence.taking.remove(lawrence.taking[0])
    db.commit()
    assert len(course1.students) == 0
    assert len(db.query(StudentAssociation).all()) == relation_count - 1
    course1.students.append(lawrence)
    db.commit()
    assert len(db.query(StudentAssociation).all()) == relation_count
    assert len(lawrence.taking) == 1
    # Check data
    association = StudentAssociation.find(db, lawrence, course1)
    assert association.first_name is None
    assert association.last_name is None
    assert association.email is None
    association.first_name = 'Lawrence.first_name'
    association.last_name = 'Lawrence.last_name'
    association.email = 'Lawrence.email'
    db.commit()
    association = StudentAssociation.find(db, lawrence, course1)
    assert association.first_name == 'Lawrence.first_name'
    assert association.last_name == 'Lawrence.last_name'
    assert association.email == 'Lawrence.email'


def test_notimpl():
    'Test not implemented functions'
    global Session
    db = Session()
    try:
        db.query(User).first().delete(db)
    except NotImplementedError:
        pass


def test_print():
    'Test __str__ functions'
    global Session
    db = Session()
    clear_db(db, None)
    init_db(db, test_storage)
    assert str(db.query(User).first()).startswith('<User')
    assert str(db.query(Course).first()).startswith('<Course')
    assert str(db.query(Assignment).first()).startswith('<Assignment')
    assert str(db.query(Submission).first()).startswith('<Submission')
    assert str(db.query(File).first()).startswith('<File')


def test_remove_course():
    global Session
    db = Session()
    for course in db.query(Course).all():
        course.delete(db)
    a = dump_db(db)
    assert list(dump_db(db)) == ['users']


def test_jhub_user():
    global Session
    db = Session()
    user_model = {'name': 'eric'}
    assert str(User.from_jupyterhub_user(user_model, db)) == '<User eric>'
    user_model = {'name': 'new_usr'}
    assert str(User.from_jupyterhub_user(user_model, db)) == '<User new_usr>'


def test_clean():
    'Clean test space'
    global Session
    db = Session()
    clear_db(db, test_storage)



import setuptools
import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ngshare",
    version=get_version('ngshare/version.py'),
    author="Team KALE",
    author_email="ercli@ucdavis.edu",
    description="nbgrader sharing service",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lxylxy123456/ngshare",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.6',
    install_requires=[
        'SQLAlchemy>=1.3.12',
        'alembic>=1.3.2',
        'tornado>=6.0.3',
        'jupyterhub>=1.1.0',
    ],
)
